-- ILLiad Reshare

require("AtlasHelpers");
require("JsonParser");
require("JsonWriter");
require("Utilities");
require("TagProcessor");

luanet.load_assembly("System");
luanet.load_assembly("System.Data");
luanet.load_assembly("log4net");

local types = {};
types["System.Type"] = luanet.import_type("System.Type");
types["System.String"] = luanet.import_type("System.String");
types["System.Array"] = luanet.import_type("System.Array");
types["System.Activator"] = luanet.import_type("System.Activator");
types["System.Reflection.Assembly"] = luanet.import_type("System.Reflection.Assembly");
types["SqlDbType"] = luanet.import_type("System.Data.SqlDbType");
types["System.DateTimeOffset"] = luanet.import_type("System.DateTimeOffset");

-- Import the LogManager type
types["log4net.LogManager"] = luanet.import_type("log4net.LogManager");

-- Create a logger
local log = types["log4net.LogManager"].GetLogger("Reshare");

local reshareAssembly = types["System.Reflection.Assembly"].LoadFile(AddonInfo.Directory .. "\\AtlasSystems.Reshare.dll");
types["AtlasSystems.Reshare.ReshareRequestClient"] = reshareAssembly:GetType("AtlasSystems.Reshare.ReshareRequestClient");
types["AtlasSystems.Reshare.ReshareSearchClient"] = reshareAssembly:GetType("AtlasSystems.Reshare.ReshareSearchClient");
types["AtlasSystems.Reshare.Models.ReshareRequest"] = reshareAssembly:GetType("AtlasSystems.Reshare.Models.ReshareRequest");
types["AtlasSystems.Reshare.Models.ReshareSearchRequestList"] = reshareAssembly:GetType("AtlasSystems.Reshare.Models.ReshareSearchRequestList");


local Settings = {};
Settings.nvtgc = GetSetting("NVTGC");
Settings.baseUrl = GetSetting("ReshareAPIUrl")
Settings.tenant = GetSetting("ReshareTenant");
Settings.openUrlListener = GetSetting("OpenURLListenerURL");
Settings.openUrlTenant = GetSetting("OpenURLTenant");
Settings.username = GetSetting("ReshareAPIUsername");
Settings.password = GetSetting("ReshareAPIPassword");
Settings.sharedIndexUrl = GetSetting("SharedIndexAPIUrl");
Settings.sharedIndexTenant = GetSetting("SharedIndexTenant");
Settings.sharedIndexUser = GetSetting("SharedIndexUsername");
Settings.sharedIndexPassword = GetSetting("SharedIndexPassword");
Settings.searchMonitorQueue = GetSetting("SearchingMonitorQueue");
Settings.notFoundQueue = GetSetting("NotFoundQueue");
Settings.unfilledQueue = GetSetting("UnfilledQueue");
Settings.requestSentQueue = GetSetting("RequestSentQueue");
Settings.pollingInterval = GetSetting("RequestPollingFrequency");
Settings.filledStateCodes = GetSetting("ReshareFilledStatuses");
Settings.unfilledStateCodes = GetSetting("ReshareUnfilledStatuses");
Settings.patronIdField = GetSetting("PatronIDField");
Settings.pickupLocationLookupValue = GetSetting("PickupLocationLookupValue");
Settings.pickupLocationLookupMapping = {};

local requestClient;
local searchClient;

local isCurrentlyProcessing = false;
local lastPolledTime = 0;

function Init()

	local loadParams = types["System.Array"].CreateInstance(types["System.Type"].GetType("System.Object"), 1);
	loadParams[0] = AddonInfo.Directory;

	local reshareLoadMethod = types["AtlasSystems.Reshare.ReshareRequestClient"]:GetMethod("LoadDependencies");
	reshareLoadMethod:Invoke(nil, loadParams);

	local requestClientParams = { Settings.baseUrl, Settings.openUrlListener, Settings.tenant, Settings.openUrlTenant, Settings.username, Settings.password };
	requestClient = types["System.Activator"].CreateInstance(types["AtlasSystems.Reshare.ReshareRequestClient"], requestClientParams);

	local searchClientParams = { Settings.sharedIndexUrl, Settings.sharedIndexTenant, Settings.sharedIndexUser, Settings.sharedIndexPassword };
	searchClient = types["System.Activator"].CreateInstance(types["AtlasSystems.Reshare.ReshareSearchClient"], searchClientParams);

	BuildPickupLocationMappings();

	log:Debug("Clients created")

    RegisterSystemEventHandler("GenerateFulfillmentData","GenerateFulfillmentData");
	RegisterSystemEventHandler("AutomationDecision","AutomationDecision");
    RegisterSystemEventHandler("SystemTimerElapsed","TimerElapsed");
end

function TimerElapsed(eventargs)
	log:Info("Reshare: Timer elapsed.")
	if (isCurrentlyProcessing == false) then
		ProcessDataContexts("TransactionStatus", Settings.searchMonitorQueue, "SearchAndRequest");

		if (os.difftime(os.time(), lastPolledTime) / 60 >= Settings.pollingInterval) then
			log:Info("Polling interval has elapsed.")
			lastPolledTime = os.time();
			local success, err = pcall(CheckReshareStatusUpdates);
			if (success == false) then
				log:Error("Error encountered polling Reshare requests by status.");
				log:Error(err.Message);
			end

		else
			log:Debug("Polling interval has not elapsed.")
		end
	end
end

-- Stage 2
function GenerateFulfillmentData(eventArgs)
	log:Info("Reshare: Generating Fulfillment Data");

	local success, result = pcall(GetItemsAvailableInSharedIndex);

	if (not success) then
		local message = "An error occurred when searching Reshare for tn " .. GetFieldValue("Transaction", "TransactionNumber");
		log:Error(message);
		ExecuteCommand("AddNote", { GetFieldValue("Transaction", "TransactionNumber"), message });
		error("An error occurred searching Reshare");
	end
	local idList = result;

	if (idList == nil) then
		log:Debug("No results found");
	else
		log:Debug(idList.Count.." results found.");
	end

	if (idList == nil or idList.Count == 0) then
		eventArgs.Available = false;
		eventArgs.Notes = "No loanable holdings found in shared index.";
	elseif (idList.Count > 1) then
		eventArgs.Available = false;
		eventArgs.Notes = "Multiple results: "..idList.Count.." records found.";
	else
		eventArgs.Available = true;
		eventArgs.Cost = true;
		eventArgs.CopyrightIncluded = true;
		eventArgs.TurnaroundTime = 0

		local customData = {};
		customData.instanceId = idList[0].Id;
		customData.title = idList[0].Title;
		eventArgs.CustomData = JsonWriter:Serialize(customData)
	end

    eventArgs.Handled = true;
end

-- Stage 4
function AutomationDecision(eventArgs)
	log:Info("Reshare: Submitting Request from DSP");

	local customData = JsonParser:ParseJSON(eventArgs.CustomData);
	local instanceId = customData.instanceId;
	local title = customData.title;
	local patronId = TagProcessor:ReplaceTags(Settings.patronIdField);
	local pickupLocationValue = TagProcessor:ReplaceTags(Settings.pickupLocationLookupValue);

	if pickupLocationValue == nil then
		pickupLocationValue = "";
	end

	if ((Settings.pickupLocationLookupMapping[pickupLocationValue] ~= nil) and (Settings.pickupLocationLookupMapping[pickupLocationValue] ~= "")) then
		pickupLocationValue = Settings.pickupLocationLookupMapping[pickupLocationValue];
	else
		log:WarnFormat("PickupLocation mapping not found for {0}", pickupLocationValue);
	end

	if (pickupLocationValue == nil or pickupLocationValue == "") then
		eventArgs.Handled = false;
	else
		CreateRequest(patronId, pickupLocationValue, instanceId, title);
		ExecuteCommand("Route", {GetFieldValue("Transaction","TransactionNumber"), Settings.requestSentQueue});

		eventArgs.Handled = true;
	end
end

function SearchAndRequest()
	--Confirm that this is a request we should be processing
	local userNvtgc = GetFieldValue("User", "NVTGC");
	if (string.lower(ILLiad.GetParentNVTGC(userNvtgc)) ~= string.lower(Settings.nvtgc)  ) then
		log:Debug("Skipping shared index search for transaction " .. GetFieldValue("Transaction", "TransactionNumber") .. " because it is not within this addon's NVTGC.");
		return;
	end

	log:Debug("Searching shared index for tn "..GetFieldValue("Transaction", "TransactionNumber"));
	isCurrentlyProcessing = true;
	local success, err = pcall(function()
		local idList = GetItemsAvailableInSharedIndex();
		if (idList == nil or idList.Count == 0 or idList.Count > 1) then
			local noteText;
			if (idList == nil or idList.Count == 0) then
				noteText = "No results found. Cannot create ReShare request.";
			else
				noteText = "More than one result. Cannot create ReShare request";
			end

			ExecuteCommand("AddNote", { GetFieldValue("Transaction", "TransactionNumber"), noteText });	
			ExecuteCommand("Route", {GetFieldValue("Transaction","TransactionNumber"), Settings.notFoundQueue});
		elseif (idList.Count == 1) then
			log:Debug("One result found for tn "..GetFieldValue("Transaction", "TransactionNumber"));
			local patronId = TagProcessor:ReplaceTags(Settings.patronIdField);

			local pickupLocationValue = TagProcessor:ReplaceTags(Settings.pickupLocationLookupValue);
			if pickupLocationValue == nil then
				pickupLocationValue = "";
			end

			if ((Settings.pickupLocationLookupMapping[pickupLocationValue] ~= nil) and (Settings.pickupLocationLookupMapping[pickupLocationValue] ~= "")) then
				pickupLocationValue = Settings.pickupLocationLookupMapping[pickupLocationValue];
			else
				log:WarnFormat("PickupLocation mapping not found for {0}", pickupLocationValue);
			end

			CreateRequest(patronId, pickupLocationValue, idList[0].Id, idList[0].Title);
			ExecuteCommand("Route", {GetFieldValue("Transaction","TransactionNumber"), Settings.requestSentQueue});
		end
	end)

    if (not success) then
        local message = "An error occurred when searching Reshare for tn " .. GetFieldValue("Transaction", "TransactionNumber");
        log:Error(message);
        ExecuteCommand("AddNote", { GetFieldValue("Transaction", "TransactionNumber"), message });
        error("An error occurred searching Reshare");
    end

	isCurrentlyProcessing = false;
end

function GetItemsAvailableInSharedIndex()
	log:Info("Reshare: Searching Shared Index");
	local idList;
	local oclcNumber = GetFieldValue("Transaction","ESPNumber");
	if (oclcNumber ~= nil and oclcNumber ~= "") then
		idList = searchClient:QueryByOclcNumber(oclcNumber);
	end
	if (idList ~= nil and idList.Count > 0) then
		return idList;
	end

	local isxn = GetFieldValue("Transaction","ISSN");
	if (isxn ~= nil and isxn ~= "") then
		idList = searchClient:QueryByIsxn(isxn);
	end
	if (idList ~= nil and idList.Count > 0) then
		return idList;
	end

	local title = GetFieldValue("Transaction","LoanTitle");
	local author = GetFieldValue("Transaction","LoanAuthor");
	local date = GetFieldValue("Transaction","LoanDate");

	if (title ~= nil and title ~= "") then
		idList = searchClient:QueryByBibInfo(title, author, date);
	end
	if (idList ~= nil and idList.Count > 0) then
		return idList;
	end

	return idList;
end

function CreateRequest(patronId, pickupLocation, instanceId, title)
	log:Debug("ReShare: Creating ReShare Request");
	local success, result = pcall(function() return requestClient:CreateReshareRequest(patronId, pickupLocation, instanceId, title) end);

	if success then
		log:Debug("ReShare: Request created. Updating ILL fields.");
		SetFieldValue("Transaction", "ILLNumber", result.Hrid);
		SetFieldValue("Transaction", "SystemID", "Reshare:"..Settings.tenant);
		SaveDataSource("Transaction");
		log:Debug("ReShare: ILL fields updated");
    else
		if (result.InnerException.Message ~= nil) then
			ExecuteCommand("AddNote", { GetFieldValue("Transaction", "TransactionNumber"), result.InnerException.Message });
		end

		local message;
		if (result.InnerException ~= nil and result.InnerException.RequestId ~= nil and result.InnerException.RequestId ~= "") then
			message = "A ReShare request was created, but the local ID could not be retrieved. ReShare request id: "..result.InnerException.RequestId;
		else
			message = "An error was encountered creating a ReShare request for transaction" .. GetFieldValue("Transaction", "TransactionNumber");
		end
		log:Error(message);
		ExecuteCommand("AddNote", { GetFieldValue("Transaction", "TransactionNumber"), message });
	end
end

function CheckReshareStatusUpdates()
	log:Info("Checking for ReShare status updates");

	local connection = CreateManagedDatabaseConnection();
	connection:Connect();

	local success, err = pcall(function()
		local unfilledStateCodes = AtlasHelpers.StringSplit(",", Settings.unfilledStateCodes);
		for _, stateCode in ipairs(unfilledStateCodes) do
			stateCode = AtlasHelpers.Trim(stateCode);
			log:DebugFormat("Checking for requests in status {0}", stateCode);
			UpdateReshareRequestsByStateCode(stateCode, Settings.unfilledQueue, connection, false);
		end

		local filledStateCodes = AtlasHelpers.StringSplit(",", Settings.filledStateCodes);
		for _, stateCode in ipairs(filledStateCodes) do
			stateCode = AtlasHelpers.Trim(stateCode);
			log:DebugFormat("Checking for requests in status {0}", stateCode);
			UpdateReshareRequestsByStateCode(stateCode, "Request Finished", connection, true)
		end
	end);

	if (not success) then
		log:Error("An error occurred checking Reshare for status updates.");
		log:Error(err);
	end

	--[SQLConnection] Close and Dispose are functionally equivalent. --MSDN
	connection:Dispose();
end


function UpdateReshareRequestsByStateCode(stateCode, newStatus, connection, includeLender)
	log:Debug("Moving "..stateCode.." requests currently in " .. Settings.requestSentQueue .. " to "..newStatus);
	if (includeLender == true) then
		log:Debug("LendingLibrary will be updated.")
	else
		log:Debug("LendingLibrary will not be updated.")
	end
	local dateTime = types["System.DateTimeOffset"].FromUnixTimeSeconds(lastPolledTime).DateTime.Date;

	local success, result = pcall(function() return requestClient:GetRequestsByStateCode(stateCode, dateTime) end );

	if (not success) then
		log:Error("An error was encountered retrieving requests by statecode "..stateCode);
		if (result.InnerException.Message ~= nil) then
			log:Error(result.InnerException.Message);
		end
		return;
	end

	local reshareRequestList = result;
	log:Debug(reshareRequestList.Results.Count .. " ReShare requests found in state " .. stateCode);
	if (reshareRequestList.Results.Count > 0) then
		local requestEnumerator = reshareRequestList.Results:GetEnumerator();
		while (requestEnumerator:MoveNext()) do
			local request = requestEnumerator.Current;

			local lender = nil;
			if (includeLender == true) then
				lender = request.ResolvedSupplier.Symbol
			end

			if (request == nil) then
				log:Warn("GetRequestsByStateCode returned null response");
			elseif (request.Hrid == nil) then
				log:Warn("GetRequestsByStateCode returned response with null Hrid");
			end;			

			UpdateByHrid(request.Hrid, Settings.requestSentQueue, newStatus, connection, lender);
		end
	end
end

-- Finds a single transaction by hrid, and routes it to newStatus if it is at currentStatus
function UpdateByHrid(hrid, currentStatus, newStatus, connection, lender)
	connection.QueryString = "SELECT TransactionNumber FROM Transactions WHERE ILLNumber = @Hrid AND TransactionStatus = @CurrentStatus;";
	connection:AddParameter('@Hrid', hrid, types.SqlDbType.NVarChar);
	connection:AddParameter('@CurrentStatus', currentStatus, types.SqlDbType.NVarChar);

	local transactions = connection:Execute();
	if transactions.Rows.Count == 1 then
		local tn = transactions.Rows:get_Item(0):get_Item(0);
		log:Debug("Routing transaction "..tn.." to "..newStatus);
		ExecuteCommand("Route", {tn, newStatus});

		-- We can't pass parameters to defined called by ProcessDataContexts,
		-- but if we define the function within this one, the desired lender value will be in scope.
		function UpdateLendingLibrary()
			if (lender ~= nil and lender ~= "") then
				log:Debug("Setting LendingLibrary to "..lender)
				SetFieldValue("Transaction", "LendingLibrary", lender);
				SaveDataSource("Transaction");
			end
		end

		ProcessDataContexts("TransactionNumber", tn, "UpdateLendingLibrary");
	elseif transactions.Rows.Count > 1 then
		log:Info("Multiple requests found with ReShare local id "..hrid". These requests will not be updated.");
	end

	connection:ClearParameters();
end

function BuildPickupLocationMappings()
	local mappings = AtlasHelpers.StringSplit(",", GetSetting("PickupLocationLookupMapping"));
	for i = 1, #mappings do
		local keyValue = {};

		log:DebugFormat("Processing PickupLocation Mapping {0}", mappings[i]);

		local equalPosition = string.find(mappings[i], "=");

		if (equalPosition == nil) then
			log:ErrorFormat("Invalid Location mapping found: {0}",  mappings[i]);
			ExecuteCommand("WriteSystemLog", "ReShare: Invalid Location mapping found: " ..  mappings[i]);
		else
			local illiadLocation = AtlasHelpers.Trim(string.sub(mappings[i], 0, equalPosition - 1));
			local reshareLocation = AtlasHelpers.Trim(string.sub(mappings[i], equalPosition + 1));
			log:DebugFormat("ILLiad Location: {0}. ReShare Location: {1}", illiadLocation, reshareLocation);
			Settings.pickupLocationLookupMapping[illiadLocation] = reshareLocation;
		end
	end
end